using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Threading; // Mutex

namespace MatrixSaver
{
	/// <summary>
	/// Summary description for MatrixForm.
	/// </summary>
	public partial class MatrixForm : Form
    {
        #region Constants

        /// <summary>
        /// Pixel rolloff buffer value. This increases the size of the
        /// resulting bitmap such that it's so many pixels higher than
        /// what is reported by the system. This allows for a smooth
        /// transition off-screen (some dual monitor systems also would
        /// leave a 2 pixel strip at the bottom...this helps to correct
        /// for that).
        /// </summary>
        private const Int32 RollOff = 5; // in pixels

        #endregion

        #region Protected properties

        /// <summary>
        /// Mouse (persisted) X/Y coordinates, used to determine
        /// mouse movement between mouse events.
        /// </summary>
        private Point _mouseXY;

        /// <summary>
        /// THE Matrix panel. This is the foundation of the screen
        /// saver (note the panel can be reused as a true Windows
        /// Forms panel...).
        /// </summary>
		private MatrixPanel pnlMatrix;

        /// <summary>
        /// Single-use mutex, since we don't want multiple copies of
        /// this application running at one time.
        /// </summary>
		private static Mutex _singleInstance = null;

        #endregion

        #region Constructor

        /// <summary>
        /// Default constructor.
        /// </summary>
        public MatrixForm()
		{
			InitializeComponent();

			// Hook the mouse movement event from the matrix panel...
			this.pnlMatrix.MouseMove += new MouseEventHandler(pnlMatrix_MouseMove);

			// Also hook the mouse down event...
			this.pnlMatrix.MouseDown += new MouseEventHandler(pnlMatrix_MouseDown);
        }

        #endregion

        #region Form load event handler

        /// <summary>
        /// Form "load" event handler.
        /// </summary>
        /// <param name="sender">Not used.</param>
        /// <param name="e">Not used.</param>
        private void MatrixForm_Load(object sender, System.EventArgs e)
		{
			// Make sure we're the only instance... We'll create a mutex, demanding
			// that we own it initially. We'll give it a name so other instances
			// of the application will try to use the same mutex. We'll then
			// check a boolean variable to see if we were granted access to the
			// mutex. If so, we proceed. If not, there is another instance
			// running and we quit.
			bool bFirst = false;
			_singleInstance = new Mutex(true, "MatrixSaver", out bFirst);

			if ( bFirst )
			{
				// We're in. Now load the options
                Options options = new Options();
                pnlMatrix.NumberOfStreams = options.NumberOfStreams;
                pnlMatrix.NumberOfTrailers = options.NumberOfTrailers;
                pnlMatrix.TrailLength = options.TrailLength;
                pnlMatrix.SpacePad = options.SpacePadding;
                
                // Then set up the monitors
                Int32 totWidth = 0;
                Int32 totHeight = 0;
                for (Int32 i = 0; i < Screen.AllScreens.Length; i++)
                {
                    // Check the dimensions of this screen
                    if ( Screen.AllScreens[i].Bounds.X == 0 )
                    {
                        // We simply compare the width of this screen to
                        // the total width we've accumulated, and if greater,
                        // use this screen's width. Otherwise, we're covered.
                        if ( Screen.AllScreens[i].Bounds.Width > totWidth )
                            totWidth = Screen.AllScreens[i].Bounds.Width;
                    } // if
                    else
                    {
                        // The widths are additive...
                        totWidth += Screen.AllScreens[i].Bounds.Width;
                    } // else

                    // Height
                    if (Screen.AllScreens[i].Bounds.Y == 0)
                    {
                        // We simply compare the height of this screen, similar
                        // to what we did for width.
                        if (Screen.AllScreens[i].Bounds.Height > totHeight)
                            totHeight = Screen.AllScreens[i].Bounds.Height;
                    } // if
                    else
                    {
                        // The heights are additive...
                        totHeight += Screen.AllScreens[i].Bounds.Height;
                    } // else
                } // for

                // Set our bounds
                Location = new Point(0, 0);
                Size = new Size(totWidth, totHeight + RollOff);
                pnlMatrix.Size = Size;

                // Two monitors, twice the number of streams to
                // make it look good...
                pnlMatrix.NumberOfStreams *= Screen.AllScreens.Length;

				// Hide the cursor
				Cursor.Hide();

				// Start the matrix
				pnlMatrix.Start();
			} // if
			else
			{
				// Not the first instance, so quit...
				_singleInstance.Close();
				Application.Exit();
			} // else
		}

		#endregion

		#region I/O event handlers

		protected override void OnMouseDown(MouseEventArgs e)
		{
			base.OnMouseDown (e);

			// Stop the Matrix
			pnlMatrix.Stop();

			// Close the form
			Close();
		}

		private void pnlMatrix_MouseDown(object sender, MouseEventArgs e)
		{
			// Delegate to the form's handler...
			OnMouseDown(e);
		}

		protected override void OnMouseMove(MouseEventArgs e)
		{
			base.OnMouseMove (e);

			// See if the mouse has already moved...
			if ( !_mouseXY.IsEmpty )
			{
				// It did, so see if we moved a pixel more
				if ( _mouseXY != new Point(e.X, e.Y ))
				{
					// Stop the Matrix
					pnlMatrix.Stop();

					// We did, so close
					Close();
				} // if

				// We didn't, but did we click?
				else if (e.Clicks > 0)
				{
					// Stop the Matrix
					pnlMatrix.Stop();

					// We did, so close
					Close();
				} // else if
			} // if

			// First movement, so persist point
			_mouseXY = new Point(e.X, e.Y);
		}

		private void pnlMatrix_MouseMove(object sender, MouseEventArgs e)
		{
			// Delegate to the form's handler...
			OnMouseMove(e);
		}

		protected override void OnKeyDown(KeyEventArgs e)
		{
			base.OnKeyDown (e);

			// Stop the Matrix
			pnlMatrix.Stop();

			// Close the form
			Close();
		}

		#endregion
	}
}
