using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using System.Diagnostics; // Process
using System.Runtime.InteropServices; // DllImport

namespace MatrixSaver
{
    /// <summary>
    /// The form for assigning user-specified configurable
    /// parameters.
    /// </summary>
    public partial class MatrixOptions : Form
    {
        #region Interop definitions

        /// <summary>
        /// Win32 API message WM_NULL. WPARAM and LPARAM are unused. Returns
        /// zero if application processes the message.
        /// </summary>
        private const UInt32 WM_NULL = 0x0000;

        /// <summary>
        /// "Send message timeout" blocking option. Our thread blocks while
        /// waiting for the called application to process the given Win32
        /// message.
        /// </summary>
        private const UInt32 SMTO_BLOCK = 0x0001;

        /// <summary>
        /// Sends a Win32 message to the given window with a timeout, in case
        /// the window isn't responding. 
        /// </summary>
        /// <param name="hWnd">Window handle of window to call.</param>
        /// <param name="Msg">Win32 message to process.</param>
        /// <param name="wParam">WPARAM to process.</param>
        /// <param name="lParam">LPARAM to process</param>
        /// <param name="fuFlags">Execution flags</param>
        /// <param name="uTimeout">Timeout (in milliseconds)</param>
        /// <param name="lpdwResult">Value returned from application when Win32
        /// message is processed (message dependent).</param>
        /// <returns></returns>
        [DllImport("user32.dll", CallingConvention = CallingConvention.StdCall, EntryPoint = "SendMessageTimeoutW")]
        private static extern UInt32 SendMessageTimeout(IntPtr hWnd,
                                                        UInt32 Msg,
                                                        UInt32 wParam,
                                                        UInt32 lParam,
                                                        UInt32 fuFlags,
                                                        UInt32 uTimeout,
                                                        out UInt32 lpdwResult);
        #endregion

        #region Constants

        /// <summary>
        /// The URL for the Wintellect home page.
        /// </summary>
        protected const string WintellectHomePageUrl = "http://www.wintellect.com";

        /// <summary>
        /// Period (milliseconds) used to check for the visibility
        /// of the parent window.
        /// </summary>
        protected const Int32 Period = 100;

        /// <summary>
        /// Delay for SendMessageTimeout, in milliseconds.
        /// </summary>
        protected const Int32 SMTODelay = 10;

        /// <summary>
        /// Tooltip strings.
        /// </summary>
        protected const string TTMaxStreams = "Maximum number of trails to display";
        protected const string TTMaxChars = "Number of characters forming the trail";
        protected const string TTMaxTrailers = "Number of characters displayed in the trail before erasing begins";
        protected const string TTSpacePad = "Number of random characters used to increase or decrease the trail length";

        #endregion

        #region Protected properties

        /// <summary>
        /// The Options object we're tied to for loading/saving
        /// screen saver options.
        /// </summary>
        protected Options _options = null;

        /// <summary>
        /// The parent window we're slaved to.
        /// </summary>
        protected IntPtr _hwndParent = IntPtr.Zero;

        /// <summary>
        /// The timer we'll use to periodically test the visibility
        /// of our parent (if we are slaved to a parent).
        /// </summary>
        protected System.Windows.Forms.Timer _timer = null;

        /// <summary>
        /// Tooltips, for user assistance with the meaning of the
        /// options.
        /// </summary>
        protected ToolTip _ttLblStreams = null;
        protected ToolTip _ttNumStreams = null;
        protected ToolTip _ttLblTrailers = null;
        protected ToolTip _ttNumTrailers = null;
        protected ToolTip _ttLblTrailerLen = null;
        protected ToolTip _ttTrailerLen = null;
        protected ToolTip _ttLblSpacePad = null;
        protected ToolTip _ttSpacePad = null;

        #endregion

        #region Constructors

        /// <summary>
        /// Default constructor.
        /// </summary>
        public MatrixOptions()
        {
            InitializeComponent();

            // Create and assign the tooltips
            this._ttLblStreams = new ToolTip();
            this._ttLblStreams.SetToolTip(this.lblNumStreams, TTMaxStreams);
            this._ttNumStreams = new ToolTip();
            this._ttNumStreams.SetToolTip(this.cmbNumStreams, TTMaxStreams);

            this._ttLblTrailers = new ToolTip();
            this._ttLblTrailers.SetToolTip(this.lblTrailers, TTMaxChars);
            this._ttNumTrailers = new ToolTip();
            this._ttNumStreams.SetToolTip(this.cmbNumTrailers, TTMaxChars);

            this._ttLblTrailerLen = new ToolTip();
            this._ttLblTrailers.SetToolTip(this.lblTrailLen, TTMaxTrailers);
            this._ttTrailerLen = new ToolTip();
            this._ttNumStreams.SetToolTip(this.cmbTrailLen, TTMaxTrailers);

            this._ttLblSpacePad = new ToolTip();
            this._ttLblTrailers.SetToolTip(this.lblSpacePad, TTSpacePad);
            this._ttSpacePad = new ToolTip();
            this._ttNumStreams.SetToolTip(this.cmbSpacePad, TTSpacePad);

            // Load the options
            this._options = new Options();

            // Fill the combo boxes with the options values
            AssignUIValues();
        }

        /// <summary>
        /// Owned window constructor. In this case we'll need to start
        /// a timer to periodically check for the visibility of the
        /// parent. If/when the parent becomes non-visible, we'll tear
        /// ourselves down. This mimics the old-style Windows screen
        /// saver behavior where we slaved the options dialog to the
        /// parent through windows management API functions. This is
        /// difficult to do in modern versions of Windows due to
        /// increased security models. This is a whole lot easier!
        /// </summary>
        /// <param name="hwndParent">The window handle of the parent.</param>
        public MatrixOptions(IntPtr hwndParent) : this()
        {
            // Save the parent window handle
            this._hwndParent = hwndParent;

            // Create a timer. We'll set the timer parameters
            // and initiate cycles when the form loads.
            _timer = new System.Windows.Forms.Timer();
        }

        #endregion

        #region Event handlers

        /// <summary>
        /// "OK" (accept) button handler.
        /// </summary>
        /// <param name="sender">Not used.</param>
        /// <param name="e">Not used.</param>
        private void cmdOK_Click(object sender, EventArgs e)
        {
            // Save the options
            RetrieveUIValues();

            // Close...
            this.DialogResult = DialogResult.OK;
        }

        /// <summary>
        /// "Cancel" (cancel) button handler.
        /// </summary>
        /// <param name="sender">Not used.</param>
        /// <param name="e">Not used.</param>
        private void cmdCancel_Click(object sender, EventArgs e)
        {
            // Just close...
            this.DialogResult = DialogResult.Cancel;
        }

        /// <summary>
        /// "Wintellect" hyperlink button handler.
        /// </summary>
        /// <param name="sender">Not used.</param>
        /// <param name="e">Not used.</param>
        private void llWintellect_LinkClicked(object sender, LinkLabelLinkClickedEventArgs e)
        {
            // Bring up the Wintellect page...this process is slow for some reason, so
            // show the wait cursor...
            lblLoading.Show();
            Application.DoEvents(); // allows the label to fire its paint method and show...
            this.Cursor = Cursors.WaitCursor;
            try
            {
                // Launch the default browser...
                Process.Start(WintellectHomePageUrl);
            } // try
            catch
            {
                // Tell user...
                MessageBox.Show("Unable to launch default Web browser...", "The Matrix", MessageBoxButtons.OK, MessageBoxIcon.Error);
            } // catch
            lblLoading.Hide();
            this.Cursor = Cursors.Arrow;
        }

        /// <summary>
        /// Form "on load" handler.
        /// </summary>
        /// <param name="sender">Not used.</param>
        /// <param name="e">Not used.</param>
        private void MatrixOptions_Load(object sender, EventArgs e)
        {
            // Start the timer
            if (_timer != null)
            {
                // Start the timer
                _timer.Interval = Period;
                _timer.Tick += new EventHandler(Timer_Tick);
                _timer.Start();
            } // if
        }

        /// <summary>
        /// Timer tick handler.
        /// </summary>
        /// <param name="sender">Not used.</param>
        /// <param name="e">Not used.</param>
        void Timer_Tick(object sender, EventArgs e)
        {
            // Stop the timer for the moment
            _timer.Stop();

            // Check for the availability of our parent. We'll send a null message
            // to the parent window. If the parent responds, we restart the timer
            // and continue. If the message times out, our assumption is the
            // parent window is now gone (i.e.: the user canceled the desktop
            // properties dialog without first closing our options dialog). In that
            // case, we close ourselves with a cancel operation.
            UInt32 lpdwResult;
            if (SendMessageTimeout(_hwndParent, WM_NULL, 0, 0, SMTO_BLOCK, SMTODelay, out lpdwResult) == 0)
            {
                // Parent isn't responding, so we'll close ourselves
                this.DialogResult = DialogResult.Cancel;
            } // if
            else
            {
                // Parent is still up and running, so restart the timer
                _timer.Start();
            } // else
        }

        #endregion

        #region Helper methods

        /// <summary>
        /// Pulls values from the options object and sets the
        /// UI control values accordingly.
        /// </summary>
        private void AssignUIValues()
        {
            // The options bag has integer-based values for the
            // number of streams, number of trailing characters,
            // and so forth. We'll read that and do our best to
            // convert the values we see there to values we have
            // for the user to select in the various combo boxes.
            // There isn't a great way to do this...we could assign
            // keys/values to a hash table for each combo box, but
            // it's probably just as easy to loop and do string
            // comparisons. We'd miss if someone edited the options
            // XML file by hand, but we'll account for that after
            // each loop by examining the selected index. If still
            // -1, we'll pick one (the item in the zeroth position).
            //
            // Start with the number of streams
            for ( Int32 i = 0; i < this.cmbNumStreams.Items.Count; i++ )
            {
                // Check this value
                if (_options.NumberOfStreams.ToString() == this.cmbNumStreams.Items[i].ToString())
                {
                    // Found it...
                    this.cmbNumStreams.SelectedItem = this.cmbNumStreams.Items[i];
                    break;
                } // if
            } // foreach

            // Double-check the number of streams
            if (this.cmbNumStreams.SelectedIndex == -1) this.cmbNumStreams.SelectedItem = this.cmbNumStreams.Items[0];

            // Next the number of trailing characters
            for (Int32 i = 0; i < this.cmbNumTrailers.Items.Count; i++)
            {
                // Check this value
                if (_options.NumberOfTrailers.ToString() == this.cmbNumTrailers.Items[i].ToString())
                {
                    // Found it...
                    this.cmbNumTrailers.SelectedItem = this.cmbNumTrailers.Items[i];
                    break;
                } // if
            } // foreach

            // Double-check the number of trailers
            if (this.cmbNumTrailers.SelectedIndex == -1) this.cmbNumTrailers.SelectedItem = this.cmbNumTrailers.Items[0];

            // Next the trail length
            for (Int32 i = 0; i < this.cmbTrailLen.Items.Count; i++)
            {
                // Check this value
                if (_options.TrailLength.ToString() == this.cmbTrailLen.Items[i].ToString())
                {
                    // Found it...
                    this.cmbTrailLen.SelectedItem = this.cmbTrailLen.Items[i];
                    break;
                } // if
            } // foreach

            // Double-check the trail length
            if (this.cmbTrailLen.SelectedIndex == -1) this.cmbTrailLen.SelectedItem = this.cmbTrailLen.Items[0];

            // Finally the space padding
            for (Int32 i = 0; i < this.cmbSpacePad.Items.Count; i++)
            {
                // Check this value
                if (_options.SpacePadding.ToString() == this.cmbSpacePad.Items[i].ToString())
                {
                    // Found it...
                    this.cmbSpacePad.SelectedItem = this.cmbSpacePad.Items[i];
                    break;
                } // if
            } // foreach

            // Double-check the space padding
            if (this.cmbSpacePad.SelectedIndex == -1) this.cmbSpacePad.SelectedItem = this.cmbSpacePad.Items[0];
        }

        /// <summary>
        /// Pulls values from UI controls and stuffs them into
        /// the Options object for persistence.
        /// </summary>
        private void RetrieveUIValues()
        {
            // Here we'll just access the various combo boxes and
            // pull their selected values. The values should all parse
            // into integers since we built the lists, and they contained
            // only integers when we built them...
            _options.NumberOfStreams = Int32.Parse(this.cmbNumStreams.Items[this.cmbNumStreams.SelectedIndex].ToString());
            _options.NumberOfTrailers = Int32.Parse(this.cmbNumTrailers.Items[this.cmbNumTrailers.SelectedIndex].ToString());
            _options.TrailLength = Int32.Parse(this.cmbTrailLen.Items[this.cmbTrailLen.SelectedIndex].ToString());
            _options.SpacePadding = Int32.Parse(this.cmbSpacePad.Items[this.cmbSpacePad.SelectedIndex].ToString());
        }

        #endregion
    }
}