using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Collections;
using System.ComponentModel;
using System.Windows.Forms;
using System.Threading;

namespace MatrixSaver
{
	/// <summary>
	/// Summary description for MatrixPanel. Note that the original
	/// algorithm for the Matrix simulation itself came from Louai
	/// Munajim from his Web site:
	/// 
	/// http://elouai.com/the_matrix_screensaver.php
	/// 
	/// I made some improvements and tweaks, and any bugs associated
	/// with my tinkering are mine and mine alone.
	/// </summary>
	public class MatrixPanel : System.Windows.Forms.Panel
	{
		#region Constants

		/// <summary>
		/// Maximum number of streams allowed.
		/// </summary>
		public const Int32 MaxStreams = 10000;

		/// <summary>
		/// The old C Runtime "MAX_RAND" value, which comes
		/// from the original Matrix stream algorithm.
		/// </summary>
		protected const Int32 MaxRand = 32767;

		/// <summary>
		/// Default number of streams.
		/// </summary>
		protected const Int32 DefaultNumberOfStreams = 1000;

		/// <summary>
		/// Default number of characters forming the stream trail.
		/// </summary>
		protected const Int32 DefaultNumberOfTrailers = 40;

		/// <summary>
		/// Default number of characters forming the stream trail
		/// before erasing begins (forms the solid part of the trail).
		/// </summary>
		protected const Int32 DefaultTrailLength = 10;

		/// <summary>
		/// Default number of random characters used to increase
		/// or decrease the tail length.
		/// </summary>
		protected const Int32 DefaultSpacePad = 30;

		/// <summary>
		/// Default delay before advancing lead stream character.
		/// </summary>
		protected const Int32 DefaultAdvanceDelay = 5;

		/// <summary>
		/// Default timer period (in milliseconds)
		/// </summary>
		protected const Int32 DefaultTimerPeriod = 50;

		#endregion

		#region Protected properties

		/// <summary>
		/// Maximum number of streams in play
		/// </summary>
		protected Int32 _numStreams = DefaultNumberOfStreams;

		/// <summary>
		/// Number of characters forming the stream trail.
		/// </summary>
		protected Int32 _numTrailers = DefaultNumberOfTrailers;

		/// <summary>
		/// Number of characters forming the stream trail before
		/// erasing begins (forms the solid part of the trail).
		/// </summary>
		protected Int32 _trailLength = DefaultTrailLength;

		/// <summary>
		/// Number of random characters used to increase
		/// or decrease the trail length.
		/// </summary>
		protected Int32 _spacePad = DefaultSpacePad;

		/// <summary>
		/// Delay before advancing lead stream character.
		/// </summary>
		protected Int32 _advanceDelay = DefaultAdvanceDelay;

		/// <summary>
		/// Timer period (in milliseconds).
		/// </summary>
		protected Int32 _period = DefaultTimerPeriod;

		/// <summary>
		/// Color storage
		/// </summary>
		protected Int32 _r = 150;
		protected Int32 _g = 255;
		protected Int32 _b = 100;

		/// <summary>
		///  Steam starting coordinates
		/// </summary>
		protected Int32[] _startX = null;
		protected Int32[] _startY = null;

		/// <summary>
		/// Stream speeds (will be adjusted to vary individual stream
		/// advancement velocity).
		/// </summary>
		protected Int32[] _streamSpeed = null;
		protected Int32[] _streamOrigSpeed = null;

		/// <summary>
		/// Stream active/inactive status.
		/// </summary>
		protected bool[] _streamStatus = null;

		/// <summary>
		/// Temp for initializing and maintaining streams.
		/// </summary>
		protected Int32 _streamCount = 0;

		/// <summary>
		/// Text sizing parameters (will be adjusted when
		/// a font is introduced).
		/// </summary>
		protected Int32 _textHeight = 13; // placeholders recalculated with font change
		protected Int32 _textWidth = 13; // placeholders recalculated with font change

		/// <summary>
		/// Printable stream characters (Unicode).
		/// </summary>
		protected static char[] _chars = 
		{
			'\u046c','\u029e','\u0293','\u036a','\u03d4','\u03e0','\u03e7','\u03ea','\u03f0','\u046a',	
			'\u0472','\u0463','\u0468','\u0444','\u04b0','\u04ec','\u0182','\u0163','\u01c2','\u019b',	
			'\u028a','\u0290','\u0232','\u0275','\u039e','\u03b6','\u03de','\u0416','\u03d7','\u04a6',
			'\u04dd','\u04c4','\u05d0','\u04f1','\u0e54','\u0e5a','\u0e29','\u2206','\u20aa','\u20ad',	
			'\ufb2e','\u2113','\u20ab','\u20ae','\u1fb1','\u1f73','\u1fb7','\u1f27','\u1e4c','\u0e53',
			'\u06d7','\u05e6','\u05de','\u04b4','\u0490','\u04a9','\u0409','\u0390','\u02a4','\u0224',
			'\u01aa','\u01bf','\u0126','\u00de','\u00f8','\u00e7','\u00df','\u018d','\u0195','\u0182',
			'\u04dc','\u0468','\u046d','\u0467','\u046a','\u04c1','\u04a6','\u04f4','\u04b8','\u0e1e',
			'\u0166','\u0194','\u01ab','\u01a3','\u0199','\u021c','\u021d','\u0264','\u0253','\u0255',
			'\u026e','\u022d','\u0256','\u0278','\u027b','\u0289','\u02a2','\u03f0','\u042e','\u0464',
		};

		/// <summary>
		/// Randomizer to be used.
		/// </summary>
		protected Random _rand = null;

		/// <summary>
		/// Timer to be used.
		/// </summary>
		protected System.Windows.Forms.Timer _timer = null;

		/// <summary>
		/// String format to be used.
		/// </summary>
		protected StringFormat _drawFormat = null;

		/// <summary>
		/// Bitmap onto which we'll draw the streams. 
		/// </summary>
		protected Bitmap _canvas = null;

		/// <summary>
		/// The graphics object we'll associate with the
		/// canvas bitmap.
		/// </summary>
		protected Graphics _canvasGraphics = null;

		/// <summary>
		/// Dictates whether we want exception information
		/// displayed on the screen when the saver is running.
		/// </summary>
		protected bool _displayExceptions = true;

		/// <summary>
		/// An event that holds the Dispose() method up
		/// if we're updating at the time Dispose() is
		/// called. This is critical because different
		/// threads are used, and one can call the
		/// Dispose() method while another is using the
		/// resources Dispose() is killing off. This
		/// results in bad states (believe me!).
		/// </summary>
		protected ManualResetEvent _updating = null;

		/// <summary>
		/// Running flag, used mostly for determining
		/// whether to restart the timer after setting
		/// a property.
		/// </summary>
		protected bool _running = false;

		#endregion

		#region Constructor

		/// <summary>
		/// Standard constructor.
		/// </summary>
		public MatrixPanel()
		{
			// Set double buffer style
			this.SetStyle(ControlStyles.DoubleBuffer | ControlStyles.UserPaint | ControlStyles.AllPaintingInWmPaint, true);
			this.UpdateStyles(); 

			// Create the updating event. We initialize to signaled
			// so the dispose method will run to completion. If we're
			// in the middle of updating, we'll reset the event so that
			// the dispose method will wait.
			_updating = new ManualResetEvent(true);

			// Create the timer and set initial parms
			this._timer = new System.Windows.Forms.Timer();
			_timer.Interval = _period;
			_timer.Tick += new EventHandler(OnTimer);

			// Create the background graphics object. We'll
			// dispose of this when we shut down.
			this._canvasGraphics = this.CreateGraphics();

			// Adjust the font parameters
			OnFontChanged(null);

			// Create the initial canvas
			OnSizeChanged(null);

			// Generate the random seed
			this._rand = new Random(DateTime.Now.Millisecond);

			// Reset operational matrices
			Reset();

			// Set our drawing format
			_drawFormat = new StringFormat();
			_drawFormat.FormatFlags = StringFormatFlags.FitBlackBox;
			_drawFormat.Trimming = StringTrimming.Character;
			_drawFormat.Alignment = StringAlignment.Near;
		}

		#endregion

		#region Dispose

		/// <summary>
		/// Standard Forms Dispose method. Note we'll kill off
		/// the timer, the drawing canvas, and the Graphics
		/// object we've been using.
		/// </summary>
		/// <param name="disposing">Disposal flag</param>
		protected override void Dispose(bool disposing)
		{
			if ( disposing )
			{
				// Stop the timer
				this._timer.Stop();

				// Wait for any updates to complete
				_updating.WaitOne(Timeout.Infinite, false);

				// Dispose of the event
				_updating.Close();

				// Kill the bitmap
				this._canvas.Dispose();

				// Kill the graphics
				this._canvasGraphics.Dispose();
			} // if

			base.Dispose(disposing);
		}

		#endregion

		#region Public properties

		/// <summary>
		/// Total number of streams.
		/// </summary>
		[CategoryAttribute("Runtime Parameters")]
		[DefaultValueAttribute(DefaultNumberOfStreams)]
		[DescriptionAttribute("Number of streams executing at one time.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Content | DesignerSerializationVisibility.Visible)]
		public Int32 NumberOfStreams
		{
			get { return _numStreams; }
			set
			{
				// Make sure we're stopped
				bool bRestart = _running;
				Stop();

				// Adjust boundary values
				if ( value > MaxStreams ) value = MaxStreams;
				if ( value < 1 ) value = DefaultNumberOfStreams;

				// Set value
				_numStreams = value;

				// Adjust matrices
				Reset();

				// Clear canvas (if we don't, we leave trash out there)
				Graphics g = Graphics.FromImage(_canvas);
				g.FillRectangle(Brushes.Black,0,0,_canvas.Width,_canvas.Height);
				g.Dispose();

				// Restart if necessary
				if ( bRestart ) Start();
			}
		}

		/// <summary>
		/// Total number of trailing characters.
		/// </summary>
		[CategoryAttribute("Runtime Parameters")]
		[DefaultValueAttribute(DefaultNumberOfTrailers)]
		[DescriptionAttribute("Number of characters forming the stream trail.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Content | DesignerSerializationVisibility.Visible)]
		public Int32 NumberOfTrailers
		{
			get { return _numTrailers; }
			set
			{
				// Make sure we're stopped
				bool bRestart = _running;
				Stop();

				// Adjust boundary values
				if ( value > (2*DefaultNumberOfTrailers) ) value = 2 * DefaultNumberOfTrailers;
				if ( value < 1 ) value = DefaultNumberOfTrailers;

				// Set value
				_numTrailers = value;

				// Restart if necessary
				if ( bRestart ) Start();
			}
		}

		/// <summary>
		/// Inviolate number of trailing characters before
		/// erasure begins.
		/// </summary>
		[CategoryAttribute("Runtime Parameters")]
		[DefaultValueAttribute(DefaultTrailLength)]
		[DescriptionAttribute("Number of characters forming the stream trail before erasing begins.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Content | DesignerSerializationVisibility.Visible)]
		public Int32 TrailLength
		{
			get { return _trailLength; }
			set
			{
				// Make sure we're stopped
				bool bRestart = _running;
				Stop();

				// Adjust boundary values
				if ( value > (2*DefaultTrailLength) ) value = 2 * DefaultTrailLength;
				if ( value < 1 ) value = DefaultTrailLength;
				if ( value >= _numTrailers ) value = _numTrailers / 2;

				// Set value
				_trailLength = value;

				// Restart if necessary
				if ( bRestart ) Start();
			}
		}

		/// <summary>
		/// Spacing pad, used to provide variance in the length
		/// of the lead character's trail.
		/// </summary>
		[CategoryAttribute("Runtime Parameters")]
		[DefaultValueAttribute(DefaultSpacePad)]
		[DescriptionAttribute("Number of random characters used to increase or decrease the trail length.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Content | DesignerSerializationVisibility.Visible)]
		public Int32 SpacePad
		{
			get { return _spacePad; }
			set
			{
				// Make sure we're stopped
				bool bRestart = _running;
				Stop();

				// Adjust boundary values
				if ( value > (2*DefaultSpacePad) ) value = 2 * DefaultSpacePad;
				if ( value < 1 ) value = DefaultSpacePad;

				// Set value
				_spacePad = value;

				// Restart if necessary
				if ( bRestart ) Start();
			}
		}

		/// <summary>
		/// Stream animation speed. This value is the basis
		/// for the stream's speed...it will be randomly
		/// adjusted to provide variance in stream speed
		/// across the board.
		/// </summary>
		[CategoryAttribute("Runtime Parameters")]
		[DefaultValueAttribute(DefaultAdvanceDelay)]
		[DescriptionAttribute("Delay before advancing lead stream character.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Content | DesignerSerializationVisibility.Visible)]
		public Int32 AdvanceDelay
		{
			get { return _advanceDelay; }
			set
			{
				// Make sure we're stopped
				bool bRestart = _running;
				Stop();

				// Adjust boundary values
				if ( value > (2*DefaultSpacePad) ) value = 2 * DefaultSpacePad;
				if ( value < 1 ) value = DefaultSpacePad;

				// Set value
				_advanceDelay = value;

				// Restart if necessary
				if ( bRestart ) Start();
			}
		}

		/// <summary>
		/// Timer period.
		/// </summary>
		[CategoryAttribute("Runtime Parameters")]
		[DefaultValueAttribute(DefaultTimerPeriod)]
		[DescriptionAttribute("Timer period (in milliseconds).")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Content | DesignerSerializationVisibility.Visible)]
		public Int32 TimerPeriod
		{
			get { return _period; }
			set
			{
				// Make sure we're stopped
				bool bRestart = _running;
				Stop();

				// Adjust boundary values
				if ( value > (10*DefaultTimerPeriod) ) value = 10 * DefaultTimerPeriod;
				if ( value < DefaultTimerPeriod ) value = DefaultTimerPeriod;

				// Set value
				_period = value;

				// Restart if necessary
				if ( bRestart ) Start();
			}
		}

		/// <summary>
		/// Timer period.
		/// </summary>
		[CategoryAttribute("Runtime Parameters")]
		[DefaultValueAttribute(true)]
		[DescriptionAttribute("Display exceptions on-screen when running.")]
		[DesignerSerializationVisibility(DesignerSerializationVisibility.Content | DesignerSerializationVisibility.Visible)]
		public bool DisplayExceptions
		{
			get { return _displayExceptions; }
			set { _displayExceptions = value; }
		}

		#endregion

		#region Custom methods

		/// <summary>
		/// Creates the arrays associated with the streams.
		/// We need the arrays to be large enough to contain
		/// the number of streams we wish to have available,
		/// so if we change the total number of arrays, we
		/// then also need to build new arrays (effectively
		/// resizing them to meet our need).
		/// </summary>
		protected void Reset()
		{
			_startX = new Int32[_numStreams];
			_startY = new Int32[_numStreams];
			_streamSpeed = new Int32[_numStreams];
			_streamOrigSpeed = new Int32[_numStreams];
			_streamStatus = new bool[_numStreams];
		}

		/// <summary>
		/// Starts the timer which in turn initiates
		/// animation.
		/// </summary>
		public void Start()
		{
			if ( !this.DesignMode )
			{
				// Start the timer
				_running = true;
				_timer.Start();
			} // if
		}

		/// <summary>
		/// Stops the timer, which terminates animation.
		/// </summary>
		public void Stop()
		{
			// Stop the timer
			_running = false;
			_timer.Stop();
		}

		#endregion

		#region Overrides and event handlers

		/// <summary>
		/// Resize override. We override this so we can ascertain
		/// the new size of the drawable area. We'll blow away the
		/// existing background bitmap and build a new one.
		/// </summary>
		/// <param name="e">Generic event argument (not used)</param>
		protected override void OnSizeChanged(EventArgs e)
		{
			base.OnSizeChanged (e);

			if ( _canvas != null ) _canvas.Dispose();
			_canvas = new Bitmap(this.Width,this.Height,_canvasGraphics);
		}

		/// <summary>
		/// Font override. We override this so we can capture the
		/// height and width of the new font characters.
		/// </summary>
		/// <param name="e">Generic event argument (not used)</param>
		protected override void OnFontChanged(EventArgs e)
		{
			base.OnFontChanged (e);

			// Measure the font
			SizeF size = _canvasGraphics.MeasureString("n",this.Font);
			_textHeight = Convert.ToInt32(size.Height);
			_textWidth = Convert.ToInt32(size.Width);
		}

		/// <summary>
		/// Primary timer event handler. Note it simply
		/// forces a redraw...
		/// </summary>
		/// <param name="sender">Standard event sender object (not used)</param>
		/// <param name="e">Generic event argument (not used)</param>
		protected void OnTimer(object sender, System.EventArgs e)
		{
			// Okay for exceptions to bubble up, but we certainly
			// want to set the manual reset event so that we don't
			// hang when we close...
			try
			{
				// Reset the updating event so we can't dispose
				// of ourselves prior to update completion.
				_updating.Reset();

				// Stop the timer in case the refresh action takes
				// longer than the timer period (for large screens
				// this will almost certainly be the case).
				_timer.Stop();

				// Okay to refresh
				Refresh();

				// Restart the timer
				_timer.Start();
			} // try
			finally
			{
				// Okay to dispose
				_updating.Set();
			} // finally
		}

		/// <summary>
		/// You can have any background color you choose, as long
		/// as it's black. :) 
		/// </summary>
		/// <param name="e">Generic event argument (not used)</param>
		protected override void OnBackColorChanged(EventArgs e)
		{
			// Force black
			if ( this.BackColor != Color.Black )
			{
				this.BackColor = Color.Black;

				base.OnBackColorChanged (e);
			} // if
		}

		/// <summary>
		/// Primary painting override.
		/// </summary>
		/// <param name="e">Paint event argument</param>
		protected override void OnPaint(PaintEventArgs e)
		{
			// Draw in design mode differently...
			if ( this.DesignMode )
			{
				// You might be tempted to let the Matrix run in design mode, but
				// resist the temptation. The timer will only tick when you leave
				// the design window and return, so the effect is lost. So we'll
				// just print a notice in the upper left corner and leave it at that.
				e.Graphics.DrawString("Matrix Simulation Panel",
					this.Font, 
					new SolidBrush(Color.FromArgb(_r,_g,_b)),
					0f,
					0f,
					StringFormat.GenericDefault);
			} // if
			else
			{
				try
				{
					// Draw the streams. We use a separate canvas image
					// because if we didn't, the trails would be lost
					// (you would only see the two characters drawn in
					// DisplyStreams...all trailing characters would be
					// lost since they aren't carried forward between
					// refreshes any other way). So we'll keep the
					// background image in memory, draw on that, and 
					// then copy the image to the screen. In a sense
					// we've double-buffered the double-buffer. :)
					lock (this)
					{
						CreateDestroyStreams();
						UpdateStreams();
						using ( Graphics g = Graphics.FromImage(_canvas) )
						{
							DisplayStreams(g);
						} // using
						e.Graphics.DrawImage(_canvas,0,0,_canvas.Width,_canvas.Height);
					} // lock
				} // try
				catch (Exception ex)
				{
					// Stop the timer...
					Stop();

					// Record an "oops" message to the screen...
					e.Graphics.DrawString("Error in the Matrix, the machines have lost...",
						this.Font,
						new SolidBrush(Color.FromArgb(_r,_g,_b)),
						0f,
						0f,
						StringFormat.GenericDefault);

					if ( _displayExceptions )
					{
						// If requested, place error information on screen...
						e.Graphics.DrawString(ex.ToString(),
							this.Font,
							new SolidBrush(Color.FromArgb(_r,_g,_b)),
							0f,
							40f,
							StringFormat.GenericDefault);
					} // if

					// Better let them know what to do...
					e.Graphics.DrawString("Press any key to exit...",
						this.Font,
						new SolidBrush(Color.FromArgb(_r,_g,_b)),
						0f,
						150f,
						StringFormat.GenericDefault);
				} // catch
			} // else
		}

		#endregion

		#region Matrix stream methods

		/// <summary>
		/// Rage through the stream arrays and maintain active
		/// status or re-initialize as required.
		/// </summary>
		void CreateDestroyStreams()
		{
			// If we have streams in a reset state, initialize them.
			if ( _streamCount < _numStreams )
			{
				// We do have some in a reset state, so look for
				// those streams for re-initialization.
				for(Int32 i = 0; i < _numStreams; i++)
				{
					// Streams with a stream status of false are
					// in a reset state. If we find one, do the
					// re-initialization.
					if ( !_streamStatus[i] )
					{
						// Found one...
						++_streamCount; // stream active
						_streamStatus[i] = true; // stream active
						_startY[i] = 0; // reset starting Y coordinate

						// This may take some explanation... Random.Next(limit)
						// returns a value from 0-(limit-1), so if we then divide
						// by the limit, we're getting a number back that
						// ranges from 0 - 1.
						//
						// Dividing the window width by the width of a text
						// character gives us the total number of columns
						// available for streaming. If we multiply the total
						// number of columns by the value we obtained above
						// (the random value ranging from 0 - 1), we select
						// a column from 0 - total number of columns. 
						//
						// Multiplying the column we've selected by the width
						// of a text character gives us a window coordinate we
						// can use for painting the stream.
						_startX[i] = ((_rand.Next(MaxRand)*(this.Width/_textWidth))/MaxRand)*_textWidth;

						// Using the same idea, we'll generate a random number
						// between 0 - 1 and select a stream speed. The speed
						// dictates how fast the stream moves vertically, top
						// to bottom.
						_streamOrigSpeed[i] = (_rand.Next(MaxRand)*_advanceDelay)/MaxRand;

						// Make a copy of the speed value we calculate. As we use
						// this stream, we'll slow it down. Once we get to a zero
						// speed, we'll recover the value we calculated here.
						_streamSpeed[i] = _streamOrigSpeed[i];
						break;
					} // if
				} // for
			} // if

			// Check for the need to reset a stream
			for (Int32 i = 0; i <_numStreams; i++)
			{
				// Follow the lead character below the visible screen to be
				// sure the trail is fully erased. Once the lead character
				// dips below the screen far enough to be sure the trail is
				// erased, reset the stream's availability for streaming anew.
				if ( _startY[i] > (this.Height + (_numTrailers * _textHeight)) )
				{
					_streamStatus[i] = false; // make available for reset
					--_streamCount; // one less stream in use
				} // if
			} // for
		}

		/// <summary>
		/// Updates the stream speed to give variance to the streams
		/// as they progress down the screen. Note the variance
		/// is exhibited when the stream is re-used, not as the
		/// stream is in-use.
		/// </summary>
		void UpdateStreams()
		{
			for (Int32 i = 0; i < _numStreams; i++)
			{
				// For each active stream, we'll adjust its
				// speed to give variance to the movement
				// on the screen.
				if ( _streamStatus[i] )
				{
					// If the stream speed is zero, recover
					// the originally calculated speed.
					if ( _streamSpeed[i] == 0 )
					{
						_startY[i] += _textHeight;
						_streamSpeed[i] = _streamOrigSpeed[i];
					} // if
					else
					{
						// Slow it down a notch.
						--_streamSpeed[i];
					} // else
				} // if
			} // for
		}

		/// <summary>
		/// Draws the individual streams on the canvas we created when
		/// we sized the window. Three sets of characters are drawn:
		/// 
		/// 0: Leading character (brightest color)
		/// 1: First trailing character (subdued color)
		/// 2: Trail erasure (black to eliminate trail over time)
		/// </summary>
		/// <param name="gfx">Graphics object used for drawing.</param>
		void DisplayStreams(Graphics gfx)
		{
            // Pre-allocate some memory
			float x; // X coordinate for drawn character
			float y; // Y coordinate for drawn character
            Int32 incR; // color increments for color differentiation
            Int32 incG;
            Int32 incB;
            RectangleF rect; // Character rectangle
	
			// For each stream...
			for (Int32 i = 0; i < _numStreams; i++)
			{
				// ...for each active stream...
				if ( _streamStatus[i] )
				{
					// ...pull the starting coordinates.
					x = (float)_startX[i];
					y = (float)_startY[i];

					// Cycle through the drawing states
					for (Int32 j = 0; j < 3; j++)
					{
						if ( j == 0 )
						{
							// Draw the lead character in the brightest color
							incR = _r/(_advanceDelay+1);
							incG = _g/(_advanceDelay+1);
							incB = _b/(_advanceDelay+1);

							Brush brush = new SolidBrush(Color.FromArgb(_r-_streamOrigSpeed[i]*incR,_g-_streamOrigSpeed[i]*incG,_b-_streamOrigSpeed[i]*incB));
							rect = new RectangleF(x,y,(float)_textWidth,(float)_textHeight);
							gfx.DrawString(String.Format("{0}",(_chars[(_rand.Next(MaxRand)*_chars.Length)/MaxRand])),this.Font,brush,rect,_drawFormat);
							brush.Dispose();
						} // if
						else if ( j == 1 )
						{
							// Draw the first trailing character in a subdued color
							incR = _r/3/(_advanceDelay+1);
							incG = _g/3/(_advanceDelay+1);
							incB = _b/3/(_advanceDelay+1);

							Brush brush = Brushes.Black; // no disposal required
							rect = new RectangleF(x,y-(float)_textHeight,(float)_textWidth,(float)_textHeight);
							gfx.FillRectangle(brush,rect);

							brush = new SolidBrush(Color.FromArgb(_r/3-_streamOrigSpeed[i]*incR,_g/3-_streamOrigSpeed[i]*incG,_b/3-_streamOrigSpeed[i]*incB));
							gfx.DrawString(String.Format("{0}",(_chars[(_rand.Next(MaxRand)*_chars.Length)/MaxRand])),this.Font,brush,rect,_drawFormat);
							brush.Dispose();
						} // else if
						else
						{
							// Erase additional trailing characters if possible/necessary
							Brush brush = Brushes.Black; // no disposal required

							rect = new RectangleF(x,y-(float)((((_rand.Next(MaxRand)*_spacePad)/MaxRand)+_trailLength)*_textHeight),(float)_textWidth,(float)_textHeight);
							gfx.FillRectangle(brush,rect);

							// The following was in the original algorithm, but I
							// removed it. Adding it in causes characters in the
							// trail to be erased more quickly, and I find I prefer
							// more characters left showing. They all eventually are
							// erased anyway...
							//
							// rect = new RectangleF(x,y-(float)(_numTrailers*_textHeight),(float)_textWidth,(float)_textHeight);
							// gfx.FillRectangle(brush,rect);
						} // else
					} // for
				} // if

				// Since this loop might take awhile, allow events to
				// process (such as mouse and key events). If we don't
                // do this, the panel will (could) take a long time to
                // respond to the user input (it appears to be locked).
				Application.DoEvents();
			} // for
		}

		#endregion
	}
}
