using System;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Diagnostics;
using System.Windows.Forms;
using System.Threading;
using System.Runtime.InteropServices; // DllImport

namespace MatrixSaver
{
	/// <summary>
	/// The MatrixPreviewWindow class implements the painting
	/// functions of the MatrixPanel, with the goal being the
	/// ability to produce a mini-preview in the desktop
	/// screensavers tab (the screensaver will appear in the
	/// small monitor image rendered there). The window handle
	/// we're to use is provided on the command line, so we'll
	/// need to reconsitute that into a true handle. We'll also
	/// animate using a while loop instead of a timer (as we
	/// would normally). We'll base the while loop on the
	/// visibility of the parent window...when the user closes
	/// the window, we'll stop, no matter their selection
	/// (okay, cancel...etc.). We would love to use the functions
	/// the matrix panel already provides, but there is too
	/// much wrapped up in the Windows Forms support there
	/// to allow for a clean reuse. The use models are
	/// simply too different to implement cleanly as a single
	/// piece of code.
	/// </summary>
	public class MatrixPreviewWindow : IDisposable
	{
		#region Interop definitions

        /// <summary>
        /// RECT structure, which is expected by GetClientRect.
        /// </summary>
        [StructLayout(LayoutKind.Sequential)]
		public struct RECT
		{
			public Int32 left;
            public Int32 top;
            public Int32 right;
            public Int32 bottom;

            public RECT(Int32 l, Int32 t, Int32 r, Int32 b)
			{
				this.left = l;
                this.top = t;
                this.right = r;
                this.bottom = b;
			}
		}

        /// <summary>
        /// Retrieves client area drawing rectangle for given window.
        /// </summary>
        /// <param name="hWnd">Window handle for window to query.</param>
        /// <param name="rect">Returned rectangle value.</param>
        /// <returns>True if rectangle structure was completed, false if error.</returns>
		[DllImport("user32.dll")]
		private static extern bool GetClientRect(IntPtr hWnd, ref RECT rect);

        /// <summary>
        /// Determines if given window is visible (if so, we know the
        /// graphics surface is valid, allowing us to draw our preview).
        /// </summary>
        /// <param name="hWnd">Window handle for window to query.</param>
        /// <returns>True if window is visible, false if not.</returns>
		[DllImport("user32.dll")]
		private static extern bool IsWindowVisible(IntPtr hWnd);

		#endregion

		#region Constants

		/// <summary>
		/// Maximum number of streams allowed.
		/// </summary>
		public const Int32 MaxStreams = 10000;

		/// <summary>
		/// The old C Runtime "MAX_RAND" value, which comes
		/// from the original Matrix stream algorithm.
		/// </summary>
		protected const Int32 MaxRand = 32767;

		/// <summary>
		/// Default number of streams.
		/// </summary>
		public const Int32 DefaultNumberOfStreams = 200;

		/// <summary>
		/// Default number of characters forming the stream trail.
		/// </summary>
		public const Int32 DefaultNumberOfTrailers = 40;

		/// <summary>
		/// Default number of characters forming the stream trail
		/// before erasing begins (forms the solid part of the trail).
		/// </summary>
		public const Int32 DefaultTrailLength = 10;

		/// <summary>
		/// Default number of random characters used to increase
		/// or decrease the tail length.
		/// </summary>
		public const Int32 DefaultSpacePad = 30;

		/// <summary>
		/// Default delay before advancing lead stream character.
		/// </summary>
		public const Int32 DefaultAdvanceDelay = 5;

		/// <summary>
		/// Default timer period (in milliseconds)
		/// </summary>
		public const Int32 DefaultTimerPeriod = 100;

		#endregion

		#region Protected properties

		/// <summary>
		/// Maximum number of streams in play
		/// </summary>
		protected Int32 _numStreams = DefaultNumberOfStreams;

		/// <summary>
		/// Number of characters forming the stream trail.
		/// </summary>
		protected Int32 _numTrailers = DefaultNumberOfTrailers;

		/// <summary>
		/// Number of characters forming the stream trail before
		/// erasing begins (forms the solid part of the trail).
		/// </summary>
		protected Int32 _trailLength = DefaultTrailLength;

		/// <summary>
		/// Number of random characters used to increase
		/// or decrease the trail length.
		/// </summary>
		protected Int32 _spacePad = DefaultSpacePad;

		/// <summary>
		/// Delay before advancing lead stream character.
		/// </summary>
		protected Int32 _advanceDelay = DefaultAdvanceDelay;

		/// <summary>
		/// Timer period (in milliseconds).
		/// </summary>
		protected Int32 _period = DefaultTimerPeriod;

		/// <summary>
		/// Color storage
		/// </summary>
		protected Int32 _r = 150;
		protected Int32 _g = 255;
		protected Int32 _b = 100;

		/// <summary>
		///  Steam starting coordinates
		/// </summary>
		protected Int32[] _startX = null;
		protected Int32[] _startY = null;

		/// <summary>
		/// Stream speeds (will be adjusted to vary individual stream
		/// advancement velocity).
		/// </summary>
		protected Int32[] _streamSpeed = null;
		protected Int32[] _streamOrigSpeed = null;

		/// <summary>
		/// Stream active/inactive status.
		/// </summary>
		protected bool[] _streamStatus = null;

		/// <summary>
		/// Temp for initializing and maintaining streams.
		/// </summary>
		protected Int32 _streamCount = 0;

		/// <summary>
		/// Text sizing parameters (will be adjusted when
		/// a font is introduced).
		/// </summary>
		protected Int32 _textHeight = 13; // placeholders recalculated with font change
		protected Int32 _textWidth = 13; // placeholders recalculated with font change

		/// <summary>
		/// Printable stream characters. The true screen saver uses more interesting
		/// characters in the character set, but the preview window is too small to
		/// show them. The traditional ASCII characters are much more interesting to
		/// see in the preview window since they're less complex, visually-speaking.
		/// </summary>
		protected static char[] _chars = 
		{
			'0','1','2','3','4','5','6','7','8','9',	
			'a','b','c','d','e','f','g','h','i','j',	
			'k','l','m','n','o','p','q','r','s','t',
			'u','v','w','x','y','z','A','B','C','D',	
			'E','F','G','H','I','J','K','L','M','N',
			'O','P','Q','R','S','T','U','V','W','X',
			'Y','Z','!','@','#','$','%','^','&','*',
			'+','=','[',']','{','}','/','?','<','>',
		};

		/// <summary>
		/// Randomizer to be used.
		/// </summary>
		protected Random _rand = null;

		/// <summary>
		/// String format to be used.
		/// </summary>
		protected StringFormat _drawFormat = null;

		/// <summary>
		/// Bitmap onto which we'll draw the streams. 
		/// </summary>
		protected Bitmap _canvas = null;

		/// <summary>
		/// The graphics object we'll associate with the
		/// canvas bitmap.
		/// </summary>
		protected Graphics _canvasGraphics = null;

		/// <summary>
		/// An event that holds the Dispose() method up
		/// if we're updating at the time Dispose() is
		/// called. This is critical because different
		/// threads are used, and one can call the
		/// Dispose() method while another is using the
		/// resources Dispose() is killing off. This
		/// results in bad states (believe me!).
		/// </summary>
		protected ManualResetEvent _updating = null;

		/// <summary>
		/// The location of the upper left-hand corner
		/// of the rendering rectangle.
		/// </summary>
		protected Point _location = new Point(0, 0);

		/// <summary>
		/// The size of the rendering rectangle.
		/// </summary>
		protected Size _size = new Size(1, 1);

		/// <summary>
		/// The font we're to use.
		/// </summary>
		protected Font _font = null;

		/// <summary>
		/// The parent's screen graphics object we're to
		/// use to render the preview image.
		/// </summary>
		protected Graphics _screenGraphics = null;

		#endregion

		#region Constructor

		public MatrixPreviewWindow(Graphics screenGraphics)
		{
			// Persist the screen graphics
			this._screenGraphics = screenGraphics;

			// Create the updating event. We initialize to signaled
			// so the dispose method will run to completion. If we're
			// in the middle of updating, we'll reset the event so that
			// the dispose method will wait.
			_updating = new ManualResetEvent(true);

			// Generate the random seed
			this._rand = new Random(DateTime.Now.Millisecond);

			// Create the initial canvas
			this._canvas = new Bitmap(1,1,this._screenGraphics);

			// Create the background graphics object. We'll
			// dispose of this when we shut down.
			this._canvasGraphics = Graphics.FromImage(_canvas);

			// Establish a default font
			this.Font = new System.Drawing.Font("Microsoft Sans Serif", 8.25F, System.Drawing.FontStyle.Bold, System.Drawing.GraphicsUnit.Point, ((System.Byte)(0)));

			// Reset operational matrices
			Reset();

			// Set our drawing format
			_drawFormat = new StringFormat();
			_drawFormat.FormatFlags = StringFormatFlags.FitBlackBox;
			_drawFormat.Trimming = StringTrimming.Character;
			_drawFormat.Alignment = StringAlignment.Near;
		}

		#endregion

		#region IDispose

		/// <summary>
		/// Standard Forms Dispose method. Note we'll kill off
		/// the timer, the drawing canvas, and the Graphics
		/// object we've been using.
		/// </summary>
		public void Dispose()
		{
			// Wait for any updates to complete. Note that the
			// true screen saver can afford to wait an infinite
			// amount of time since we're working totally 
			// managed. However, here we're asked to terminate
			// by an outside process, of which we have no
			// control. So, we'll wait a specified period of
			// time before disposing, just to be sure we actually
			// do dispose of our resources and terminate.
			_updating.WaitOne(500, false);

			// Dispose of the event
			_updating.Close();

			// Kill the bitmap
			this._canvas.Dispose();

			// Kill the graphics
			this._canvasGraphics.Dispose();
		}

		#endregion

		#region Run

		public void Run(IntPtr hwndParent)
		{
			try
			{
				// We'll need to know how large the preview area
				// is supposed to be so we can generate the proper
				// canvas bitmap.
				RECT rect = new RECT();
				if ( GetClientRect(hwndParent, ref rect) )
				{
					// Place the drawing rectangle and set some
					// other parameters.
					Location = new Point(rect.left, rect.top);
					Size = new Size(rect.right - rect.left, rect.bottom - rect.top);
					Font = new Font("Microsoft Sans Serif", 2.75f,
                        System.Drawing.FontStyle.Bold,
                        System.Drawing.GraphicsUnit.Point,
                        ((System.Byte)(0)));

					// Since we'll break our event loop when our parent
					// window becomes not visible, we have to make sure
					// it's actually visible before we start. There is
					// a delay in the system from the time we're told to
					// display the preview and the time the preview window
					// becomes visible. So we'll wait here until it
					// becomes visible. To prevent an infinite loop, we'll
					// wait a predetremined amount of time (20 seconds)
					// for the preview window to become active. If for
					// some reason it isn't activated in time, we'll
					// break the loop and return, resulting in no
					// preview (but also no infinite loop situation
					// either).
					DateTime dtTimeout = DateTime.Now.AddSeconds(20.0);
					while ( !IsWindowVisible(hwndParent) )
					{
						// While we wait, process events...
						Application.DoEvents();

						// Now check the timeout
						if ( DateTime.Now > dtTimeout )
						{
							// Hmm...some problem, so end the loop and
							// quit...
							return;
						} // if
					} // while

					// Now the parent window is visible, so initiate
					// graphics ops. While the window is visible, we'll
					// run our special previewer object...
					while ( IsWindowVisible(hwndParent) )
					{
						try
						{
							// Delay awhile...
							System.Threading.Thread.Sleep(_period);

							// Reset the updating event so we can't dispose
							// of ourselves prior to update completion.
							_updating.Reset();

							// ...and update the screen.
							Refresh();
						} // try
						catch
						{
							// Some error, so simply break from
							// loop and clean up.
							break;
						} // catch
						finally
						{
							// Okay to dispose
							_updating.Set();
						} // finally

						// Keep processing events while we wait
						Application.DoEvents();
					} // while
				} // if
			} // try
			catch
			{
				// Do nothing...
			} // catch
		}

		#endregion

		#region Reset

		/// <summary>
		/// Creates the arrays associated with the streams.
		/// We need the arrays to be large enough to contain
		/// the number of streams we wish to have available,
		/// so if we change the total number of arrays, we
		/// then also need to build new arrays (effectively
		/// resizing them to meet our need).
		/// </summary>
		protected void Reset()
		{
			_startX = new Int32[_numStreams];
			_startY = new Int32[_numStreams];
			_streamSpeed = new Int32[_numStreams];
			_streamOrigSpeed = new Int32[_numStreams];
			_streamStatus = new bool[_numStreams];
		}

		#endregion

		#region Refresh

		/// <summary>
		/// Primary painting override.
		/// </summary>
		public void Refresh()
		{
			if ( _screenGraphics == null ) return;

			try
			{
				// Draw the streams. We use a separate canvas image
				// because if we didn't, the trails would be lost
				// (you would only see the two characters drawn in
				// DisplyStreams...all trailing characters would be
				// lost since they aren't carried forward between
				// refreshes any other way). So we'll keep the
				// background image in memory, draw on that, and 
				// then copy the image to the screen. In a sense
				// we've double-buffered the double-buffer. :)
				CreateDestroyStreams();
				UpdateStreams();
				using ( Graphics g = Graphics.FromImage(_canvas) )
				{
					DisplayStreams(g);
				} // using
				_screenGraphics.DrawImage(_canvas,_location.X,_location.Y,_canvas.Width,_canvas.Height);
			} // try
			catch
			{
				// Just record a message to the screen...
				_screenGraphics.DrawString("Error in the Matrix, the machines have lost...",this.Font,new SolidBrush(Color.FromArgb(_r,_g,_b)),0f,0f,StringFormat.GenericDefault);
			} // catch
		}

		#endregion

		#region Public properties

		/// <summary>
		/// Size property.
		/// </summary>
		public Size Size
		{
			get { return _size; }
			set
			{
				// Pull new size
				_size = value;

				// Create new canvas
				if ( _canvas != null ) _canvas.Dispose();
				_canvas = new Bitmap(_size.Width,_size.Height,_canvasGraphics);
			}
		}

		/// <summary>
		/// Font property.
		/// </summary>
		public Font Font
		{
			get { return _font; }
			set
			{
				// Dispose of old font
				if ( _font != null ) _font.Dispose();

				// Save the font
				_font = value;

				// Measure the font
				SizeF size = _canvasGraphics.MeasureString("N",_font);
				_textHeight = Convert.ToInt32(size.Height);
				_textWidth = Convert.ToInt32(size.Width);
			}
		}

		/// <summary>
		/// Location property.
		/// </summary>
		public Point Location
		{
			get { return _location; }
			set { _location = value; }
		}

		/// <summary>
		/// Screen graphics property. This is the graphics
		/// object associated with the parent we're to use
		/// for final rendering. The caller is responsible
		/// for proper disposal...
		/// </summary>
		public Graphics ScreenGraphics
		{
			get { return _screenGraphics; }
		}

		#endregion

		#region Matrix stream methods

		/// <summary>
		/// Rage through the stream arrays and maintain active
		/// status or re-initialize as required.
		/// </summary>
		protected void CreateDestroyStreams()
		{
			// If we have streams in a reset state, initialize them.
			if ( _streamCount < _numStreams )
			{
				// We do have some in a reset state, so look for
				// those streams for re-initialization.
				for(Int32 i = 0; i < _numStreams; i++)
				{
					// Streams with a stream status of false are
					// in a reset state. If we find one, do the
					// re-initialization.
					if ( !_streamStatus[i] )
					{
						// Found one...
						++_streamCount; // stream active
						_streamStatus[i] = true; // stream active
						_startY[i] = 0; // reset starting Y coordinate

						// This may take some explanation... Random.Next(limit)
						// returns a value from 0-(limit-1), so if we then divide
						// by the limit, we're getting a number back that
						// ranges from 0 - 1.
						//
						// Dividing the window width by the width of a text
						// character gives us the total number of columns
						// available for streaming. If we multiply the total
						// number of columns by the value we obtained above
						// (the random value ranging from 0 - 1), we select
						// a column from 0 - total number of columns. 
						//
						// Multiplying the column we've selected by the width
						// of a text character gives us a window coordinate we
						// can use for painting the stream.
						_startX[i] = ((_rand.Next(MaxRand)*(Size.Width/_textWidth))/MaxRand)*_textWidth;

						// Using the same idea, we'll generate a random number
						// between 0 - 1 and select a stream speed. The speed
						// dictates how fast the stream moves vertically, top
						// to bottom.
						_streamOrigSpeed[i] = (_rand.Next(MaxRand)*_advanceDelay)/MaxRand;

						// Make a copy of the speed value we calculate. As we use
						// this stream, we'll slow it down. Once we get to a zero
						// speed, we'll recover the value we calculated here.
						_streamSpeed[i] = _streamOrigSpeed[i];
						break;
					} // if
				} // for
			} // if

			// Check for the need to reset a stream
			for (Int32 i = 0; i <_numStreams; i++)
			{
				// Follow the lead character below the visible screen to be
				// sure the trail is fully erased. Once the lead character
				// dips below the screen far enough to be sure the trail is
				// erased, reset the stream's availability for streaming anew.
				if ( _startY[i] > (Size.Height + (_numTrailers * _textHeight)) )
				{
					_streamStatus[i] = false; // make available for reset
					--_streamCount; // one less stream in use
				} // if
			} // for
		}

		/// <summary>
		/// Updates the stream speed to give variance to the streams
		/// as they progress down the screen. Note the variance
		/// is exhibited when the stream is re-used, not as the
		/// stream is in-use.
		/// </summary>
		void UpdateStreams()
		{
			for (Int32 i = 0; i < _numStreams; i++)
			{
				// For each active stream, we'll adjust its
				// speed to give variance to the movement
				// on the screen.
				if ( _streamStatus[i] )
				{
					// If the stream speed is zero, recover
					// the originally calculated speed.
					if ( _streamSpeed[i] == 0 )
					{
						_startY[i] += _textHeight;
						_streamSpeed[i] = _streamOrigSpeed[i];
					} // if
					else
					{
						// Slow it down a notch.
						--_streamSpeed[i];
					} // else
				} // if
			} // for
		}

		/// <summary>
		/// Draws the individual streams on the canvas we created when
		/// we sized the window. Three sets of characters are drawn:
		/// 
		/// 0: Leading character (brightest color)
		/// 1: First trailing character (subdued color)
		/// 2: Trail erasure (black to eliminate trail over time)
		/// </summary>
		/// <param name="gfx">Graphics object used for drawing</param>
		void DisplayStreams(Graphics gfx)
		{
			float x; // X coordinate for drawn character
			float y; // Y coordinate for drawn character
			RectangleF rect; // Character rectangle
	
			// For each stream...
			for (Int32 i = 0; i < _numStreams; i++)
			{
				// ...for each active stream...
				if ( _streamStatus[i] )
				{
					// ...pull the starting coordinates.
					x = (float)_startX[i];
					y = (float)_startY[i];

					// Cycle through the drawing states
					for (Int32 j = 0; j < 3; j++)
					{
						if ( j==0 )
						{
							// Draw the lead character in the brightest color
							Int32 incR = _r/(_advanceDelay+1);
							Int32 incG = _g/(_advanceDelay+1);
							Int32 incB = _b/(_advanceDelay+1);

							Brush brush = new SolidBrush(Color.FromArgb(_r-_streamOrigSpeed[i]*incR,_g-_streamOrigSpeed[i]*incG,_b-_streamOrigSpeed[i]*incB));
							rect = new RectangleF(x,y,(float)_textWidth,(float)_textHeight);
							gfx.DrawString(String.Format("{0}",_chars[(_rand.Next(MaxRand)*_chars.Length)/MaxRand]),this.Font,brush,rect,_drawFormat);
							brush.Dispose();
						} // if
						else if ( j==1 )
						{
							// Draw the first trailing character in a subdued color
							Int32 incR = _r/3/(_advanceDelay+1);
							Int32 incG = _g/3/(_advanceDelay+1);
							Int32 incB = _b/3/(_advanceDelay+1);

							Brush brush = Brushes.Black; // no disposal required
							rect = new RectangleF(x,y-(float)_textHeight,(float)_textWidth,(float)_textHeight);
							gfx.FillRectangle(brush,rect);

							brush = new SolidBrush(Color.FromArgb(_r/3-_streamOrigSpeed[i]*incR,_g/3-_streamOrigSpeed[i]*incG,_b/3-_streamOrigSpeed[i]*incB));
							gfx.DrawString(String.Format("{0}",_chars[(_rand.Next(MaxRand)*_chars.Length)/MaxRand]),this.Font,brush,rect,_drawFormat);
							brush.Dispose();
						} // else if
						else
						{
							// Erase additional trailing characters if possible/necessary
							Brush brush = Brushes.Black; // no disposal required

							rect = new RectangleF(x,y-(float)((((_rand.Next(MaxRand)*_spacePad)/MaxRand)+_trailLength)*_textHeight),(float)_textWidth,(float)_textHeight);
							gfx.FillRectangle(brush,rect);

							// The following was in the original algorithm, but I
							// removed it. Adding it in causes characters in the
							// trail to be erased more quickly, and I find I prefer
							// more characters left showing. They all eventually are
							// erased anyway...
							//
							// rect = new RectangleF(x,y-(float)(_numTrailers*_textHeight),(float)_textWidth,(float)_textHeight);
							// gfx.FillRectangle(brush,rect);
						} // else
					} // for
				} // if
			} // for
		}

		#endregion
	}
}
