using System;
using System.Collections.Generic;
using System.Text;
using System.IO;
using System.IO.IsolatedStorage;
using System.Threading;
using System.Reflection;
using System.Xml;

namespace MatrixSaver
{
    /// <summary>
    /// Encapsulates reading/writing option information in an
    /// XML file to be stored in isolated storage.
    /// </summary>
    public class Options
    {
        #region Constants

        /// <summary>
        /// The options filename (will be found in isolated storage).
        /// </summary>
        protected const string ConfigFilename = "TheMatrixSS.config";

        /// <summary>
        /// The options file XML template we'll use to create a new
        /// options file if none exists.
        /// </summary>
        protected const string OptionsFileTemplate = "MatrixSaver.Properties.TheMatrixSS.xml";

        #endregion

        #region Protected properties

        /// <summary>
        /// Number of streams.
        /// </summary>
        protected Int32 _numStreams = 1000;

        /// <summary>
        /// Number of trailing characters.
        /// </summary>
        protected Int32 _numTrailers = 40;

        /// <summary>
        /// Trail length.
        /// </summary>
        protected Int32 _trailLen = 10;

        /// <summary>
        /// Space padding.
        /// </summary>
        protected Int32 _spacePad = 30;

        #endregion

        #region Constructor

        /// <summary>
        /// Default constructor.
        /// </summary>
        public Options()
        {
            // Retrieve settings/options
            LoadOptions();
        }

        #endregion

        #region Public properties

        /// <summary>
        /// Number of streams.
        /// </summary>
        public Int32 NumberOfStreams
        {
            get { return _numStreams; }
            set
            {
                // Save value
                _numStreams = value;
                PersistOptions();
            }
        }

        /// <summary>
        /// Number of trailing characters.
        /// </summary>
        public Int32 NumberOfTrailers
        {
            get { return _numTrailers; }
            set
            {
                // Save value
                _numTrailers = value;
                PersistOptions();
            }
        }

        /// <summary>
        /// Trail length.
        /// </summary>
        public Int32 TrailLength
        {
            get { return _trailLen; }
            set
            {
                // Save value
                _trailLen = value;
                PersistOptions();
            }
        }

        /// <summary>
        /// Space padding.
        /// </summary>
        public Int32 SpacePadding
        {
            get { return _spacePad; }
            set
            {
                // Save value
                _spacePad = value;
                PersistOptions();
            }
        }

        #endregion

        #region LoadOptions method

        /// <summary>
        /// Loads the options from an XML file found in isolated storage.
        /// </summary>
        private void LoadOptions()
        {
            XmlDocument xdoc = null;
            IsolatedStorageFile isoStore = IsolatedStorageFile.GetStore(IsolatedStorageScope.User | IsolatedStorageScope.Assembly | IsolatedStorageScope.Domain, null, null);
            IsolatedStorageFileStream isoStream = null;
            try
            {
                // The options file exists, so open it and pull the
                // persisted information.
                isoStream = new IsolatedStorageFileStream(ConfigFilename, FileMode.Open, FileAccess.Read, FileShare.Read);
                xdoc = new XmlDocument();
                xdoc.Load(isoStream);
            } // try
            catch (Exception)
            {
                // No options file exists, so generate an empty option file.
                isoStream = new IsolatedStorageFileStream(ConfigFilename, FileMode.CreateNew, FileAccess.Write, FileShare.None);

                // Load the raw options file resource
                try
                {
                    // Pull the resource stream
                    Stream stream = Assembly.GetExecutingAssembly().GetManifestResourceStream(OptionsFileTemplate);

                    // Write to the file
                    xdoc = new XmlDocument();
                    xdoc.Load(stream);
                    xdoc.Save(isoStream);
                } // try
                catch (Exception)
                {
                    // No problem...we'll go with the values the application was compiled
                    // against...
                    return;
                } // catch
            } // catch

            // Close the stream
            if (isoStream != null) isoStream.Close();

            // If we have options, pull them
            if (xdoc != null)
            {
                try
                {
                    // Pull the number of streams value
                    XmlAttribute attrValue = (XmlAttribute)xdoc.DocumentElement.SelectSingleNode("/TheMatrix/Options/@numStreams");
                    if (attrValue != null) Int32.TryParse(attrValue.Value, out _numStreams);

                    // Pull the number of trailing characters value
                    attrValue = (XmlAttribute)xdoc.DocumentElement.SelectSingleNode("/TheMatrix/Options/@numTrailers");
                    if (attrValue != null) Int32.TryParse(attrValue.Value, out _numTrailers);

                    // Pull the trail length value
                    attrValue = (XmlAttribute)xdoc.DocumentElement.SelectSingleNode("/TheMatrix/Options/@trailLen");
                    if (attrValue != null) Int32.TryParse(attrValue.Value, out _trailLen);

                    // Pull the spade padding value
                    attrValue = (XmlAttribute)xdoc.DocumentElement.SelectSingleNode("/TheMatrix/Options/@spacePad");
                    if (attrValue != null) Int32.TryParse(attrValue.Value, out _spacePad);
                } // try
                catch (Exception)
                {
                    // No problem...we'll go with the values the application was compiled
                    // against...
                } // catch
            } // if
        }

        #endregion

        #region PersistOptions method

        /// <summary>
        /// Stores the options in an XML file, which is then saved to
        /// isolated storage.
        /// </summary>
        private void PersistOptions()
        {
            XmlDocument xdoc = null;
            IsolatedStorageFile isoStore = IsolatedStorageFile.GetStore(IsolatedStorageScope.User | IsolatedStorageScope.Assembly | IsolatedStorageScope.Domain, null, null);
            IsolatedStorageFileStream isoStream = null;
            try
            {
                // The options file exists, so open it and assign the
                // persisted information.
                Stream stream = Assembly.GetExecutingAssembly().GetManifestResourceStream("MatrixSaver.TheMatrixSS.xml");
                xdoc = new XmlDocument();
                xdoc.Load(stream);

                // Assign the number of streams value
                XmlAttribute attrValue = (XmlAttribute)xdoc.DocumentElement.SelectSingleNode("/TheMatrix/Options/@numStreams");
                if (attrValue != null) attrValue.Value = _numStreams.ToString();

                // Assign the number of trailing characters value
                attrValue = (XmlAttribute)xdoc.DocumentElement.SelectSingleNode("/TheMatrix/Options/@numTrailers");
                if (attrValue != null) attrValue.Value = _numTrailers.ToString();

                // Assign the trail length value
                attrValue = (XmlAttribute)xdoc.DocumentElement.SelectSingleNode("/TheMatrix/Options/@trailLen");
                if (attrValue != null) attrValue.Value = _trailLen.ToString();

                // Assign the spade padding value
                attrValue = (XmlAttribute)xdoc.DocumentElement.SelectSingleNode("/TheMatrix/Options/@spacePad");
                if (attrValue != null) attrValue.Value = _spacePad.ToString();

                // Save the file
                isoStream = new IsolatedStorageFileStream(ConfigFilename, FileMode.Truncate, FileAccess.Write, FileShare.Read);
                xdoc.Save(isoStream);
                isoStream.Close();
            } // try
            catch (Exception)
            {
                // No problem...the screen saver will default next time around.
            } // catch
        }

        #endregion
    }
}
