using System;
using System.Collections.Generic;
using System.Drawing;
using System.Drawing.Drawing2D;
using System.Diagnostics;
using System.Globalization;
using System.Windows.Forms;

namespace MatrixSaver
{
    static class MatrixSaver
    {
        /// <summary>
        /// The entry point for our screensaver.
        /// </summary>
        /// <param name="args">Command line parameters/arguments.</param>
        static void Main(string[] args)
        {
            // Enable XP visual styles (primarily for options dialog)
            Application.EnableVisualStyles();
            Application.SetCompatibleTextRenderingDefault(false);

            // Pull command line arguments
            string cmd = args.Length > 0 ?
                (args[0].Length > 1 ?
                args[0].ToLower(CultureInfo.InvariantCulture).Trim().Substring(0, 2) :
                args[0].ToLower(CultureInfo.InvariantCulture).Trim().Substring(0, 1)) : "";

            // "/c" or no arguments indicates "configure." "/s" indicates
            // we're to run as a screensaver. "/p" indicates we're to run
            // as the child of the next argument (unsigned integer representing
            // the window handle of our parent)...failure to handle this
            // means we don't show up in the preview dialog on the
            // desktop.
            if (cmd == "/c" || cmd == "-c" || cmd == "c" || cmd == "")
            {
                // We could be asked to show the options dialog
                // given several command line arguments. The 
                // values we filtered for (to get to here) work
                // for all cases but one--an owner/parent. In that
                // case, the command line will appear as follows:
                //
                // /c:1269802
                //
                // where the value trailing the colon is the window
                // handle of the parent dialog. We'll need to examine
                // this case and bring up the options dialog in the
                // appropriate mode.
                MatrixOptions opts = null;
                if (args.Length > 0 && args[0].Length > 2)
                {
                    try
                    {
                        // First, shred out the HWND of the parent
                        // window.
                        char[] seps = { ':' };
                        string[] splitArgs = args[0].Split(seps);
                        if (splitArgs.Length >= 2)
                        {
                            // Looks good...we should have two parts,
                            // and we do, so assume for now the second
                            // part is the HWND.
                            UInt32 hwndParent = UInt32.Parse(splitArgs[1]);
                            opts = new MatrixOptions((IntPtr)hwndParent);
                        } // if
                        else
                        {
                            // No parent, so use default constructor
                            opts = new MatrixOptions();
                        } // else
                    } // try
                    catch
                    {
                        // Some error, so default to no parent...
                        opts = new MatrixOptions();
                    } // catch
                } // if
                else
                {
                    // No parent, so use default constructor
                    opts = new MatrixOptions();
                } // else
                if (opts != null) opts.ShowDialog();
            } // if
            else if (cmd == "/p" || cmd == "-p" || cmd == "p")
            {
                // "Preview" mode. There should be another parameter, which
                // we're to interpret as our parent window...
                if (args.Length < 2)
                    MessageBox.Show("Invalid command line. '/p' must have a parent window handle provided.",
                            "Matrix Screen Saver",
                            MessageBoxButtons.OK,
                            MessageBoxIcon.Exclamation);
                else
                {
                    try
                    {
                        // Pull the parent handle
                        UInt32 hwndParent = UInt32.Parse(args[1]);

                        // Start a new instance, with our parent providing us
                        // with a graphics surface onto which we'll paint...
                        MatrixPreviewWindow preview = null;
                        Graphics gfx = Graphics.FromHwnd((IntPtr)hwndParent);
                        if (gfx != null)
                        {
                            try
                            {
                                // Create the special preview version
                                preview = new MatrixPreviewWindow(gfx);

                                // Run the preview
                                preview.Run((IntPtr)hwndParent);
                            } // try
                            catch
                            {
                                // Do nothing...we'll just have no preview...
                            } // catch
                            finally
                            {
                                // Dispose of the preview object
                                if (preview != null) preview.Dispose();

                                // Be sure to dispose of the parent's graphics
                                // object...
                                if (gfx != null) gfx.Dispose();
                            } // finally
                        } // if
                    } // try
                    catch
                    {
                        // Do nothing...we'll just have no preview...
                    } // catch
                } // else
            } // else if
            else if (cmd == "/s" || cmd == "-s" || cmd == "s")
            {
                // Start a new instance with no parent specified...
                System.Windows.Forms.Application.Run(new MatrixForm());
            } // else if
            else
            {
                // Sorry...don't recognize the command line parameter...
                MessageBox.Show(String.Format("Invalid command line parameter, '{0}'.", cmd),
                        "Matrix Screen Saver",
                        MessageBoxButtons.OK,
                        MessageBoxIcon.Exclamation);
            } // else
        }
    }
}